#!/bin/bash

# scaler.sh
# Part of the Zerogame Project emulation suite for PS3 Linux
# By Jason Carter (jay@zerojay.com)

# Based heavily on billb's original mame-scaler.sh script.

# I feel a bit of an explanation is needed for this script since it's somewhat complex compared 
# to billb's original and it may not be as straightforward to understand as it should be.
# On the PS3, we have no access to hardware graphics acceleration. As a workaround for this,
# a program was created by miya, a Japanese PS3 Linux hacker, which allowed us to use the
# normally unused SPE cell units to upscale our emulator output to higher resolutions. All the
# emulators installed as part of Zerogame are configured to output their video at the original
# resolution the games used. When a game is selected from the Wahcade game list, this script
# decides, through various means, what scaling should be used without screwing up the aspect
# ratio and without lopping off huge parts of the image. It does the best that it can, however
# the scaler program used has some limitations in that it can only do certain scale sizes, so
# do not always expect every pixel on your screen to be used at all times.

# Command line input:
# scaler.sh [console name] [scaling type] [game name/path to game]

# [console name]
# This is the shortened Zerogame standardized console name (gbc instead of Gameboy Color, 
# for example). This allows us to special case emulators and consoles which use the same
# standard resolution and to dynamically decide what to do on a per-game basis for MAME.
# This parameter is required.

# [scaling type]
# This parameter allows the user to decide what type of scaling they would like to use; standard 
# scaling or smoothed scaling. Smoothed scaling means that the scaler will look at each pixel and
# will change the image to appear less jagged and blocky. If this parameter is omitted, standard
# scaling will be used. If fscaler is provided as the parameter passed, the scaler script will
# attempt to use the scaleXx method of smoothing. If sscaler is provided as the parameter passed,
# sai2x will be used instead.
# This parameter is optional.

# [game name/path to game]
# This parameter tells the script which game was selected by the user in Wahcade and launches
# it with the proper emulator.
# This parameter is required.

# First, determine the current display dimensions X is using as provided by xdpyinfo.
DISPLAY_DIMENSIONS=`xdpyinfo | grep dimensions | awk '{ print $2 }'`
DISPLAY_HEIGHT=`echo $DISPLAY_DIMENSIONS | cut -f2 -d x`
DISPLAY_WIDTH=`echo $DISPLAY_DIMENSIONS | cut -f1 -d x`
echo "X11 display width: $DISPLAY_WIDTH"
echo "X11 display height: $DISPLAY_HEIGHT"

# Rip out the last parameter passed and use it as the game to be launched.
eval GAMENAME=\$$#

# Determine the standard required resolution for the console/game being run.
# Most consoles have standard resolutions that their games run under, however
# because MAME emulates so many different hardware setups, it's impossible
# to select one standard resolution that works for all games. Instead, we use
# the MAME game info XML file to look up the game we want to play and grab
# the game's resolution from there and run with it.

# If the game we're running is supposed to be run by MAME...
if [ "$1" = "mame" ]; then
  EMU_BIN="mame"
  GAME_WIDTH="unknown"
  GAME_HEIGHT="unknown"
  ROTATE="unknown"
  GAME_XML=`$EMU_BIN -lx $GAMENAME | grep height | grep -v ATTLIST | column -t`
  # ...grab the game's resolution directly from the MAME game info XML output.
  for i in $GAME_XML; do
    # Is the game a vertical or horizontally oriented game?
    if [ `echo $i | cut -f1 -d =` == 'rotate' ]; then
      ROTATE=`echo $i | cut -f2 -d = | cut -f2 -d \"`
    fi
    if [ `echo $i | cut -f1 -d =` == 'width' ]; then
      GAME_WIDTH=`echo $i | cut -f2 -d = | cut -f2 -d \"`
    fi
    if [ `echo $i | cut -f1 -d =` == 'height' ]; then
      GAME_HEIGHT=`echo $i | cut -f2 -d = | cut -f2 -d \"`
    fi
  done
  # If the game is vertical, swap the reported width and height.
  # Note: Some games are reported as rotated 90 degrees and others are reported as rotated 270 degrees.
  # This fixes a major issue where many vertical games weren't being properly scaled without clipping.
  if [ $ROTATE == 90 ] || [ $ROTATE == 270 ]; then
    TEMP_HEIGHT=$GAME_HEIGHT
    GAME_HEIGHT=$GAME_WIDTH
    GAME_WIDTH=$TEMP_HEIGHT
  fi
fi

# If the game isn't a MAME game, we can more or less assume what resolution the game should be running
# at based on the most common resolutions used by the games themselves. Most of this data comes directly
# from the games themselves or from Wikipedia.
# We also use the console name to determine which emulator we should use to run the game.

if [ "$1" = "c64" ]; then
  EMU_BIN="x64"
  GAME_WIDTH="320"
  GAME_HEIGHT="200"
fi

if [ "$1" = "fds" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="256"
  GAME_HEIGHT="224"
fi

if [ "$1" = "sms" ]; then
   EMU_BIN="mednafen"
   GAME_WIDTH="256"
   GAME_HEIGHT="240"
fi

if [ "$1" = "gamegear" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="144"
fi

if [ "$1" = "gb" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="144"
fi

if [ "$1" = "gba" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="240"
  GAME_HEIGHT="160"
fi

if [ "$1" = "gbc" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="144"
fi

if [ "$1" = "genesis" ]; then
  EMU_BIN="xe"
  GAME_WIDTH="320"
  GAME_HEIGHT="224"
fi

if [ "$1" = "nes" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="256"
  GAME_HEIGHT="224"
fi

if [ "$1" = "ngpocket" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="152"
fi 

if [ "$1" = "ngpocketcolor" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="152"
fi 

if [ "$1" = "sega32x" ]; then
  EMU_BIN="xe"
  GAME_WIDTH="320"
  GAME_HEIGHT="224"
fi

if [ "$1" = "segacd" ]; then
  EMU_BIN="xe"
  GAME_WIDTH="320"
  GAME_HEIGHT="224"
fi

if [ "$1" = "tg16" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="256"
  GAME_HEIGHT="240"
fi

if [ "$1" = "snes" ]; then
  cp ~/.snes9x/snes9x-orig.xml ~/.snes9x/snes9x.xml
  EMU_BIN="snes9x-gtk"
  GAME_WIDTH="256"
  GAME_HEIGHT="224"
  if ( [[ "$GAMENAME" =~ "Secret of Mana" ]] ); then
   GAMEEXCEPTION=1
   MANAHACK=1
   echo "Secret of Mana workaround."
  fi
  if ( [[ "$GAMENAME" =~ "Seiken Densetsu 2" ]] ); then
   GAMEEXCEPTION=1
   MANAHACK=1
   echo "Secret of Mana workaround."
  fi
fi

# The Atari Lynx, Bandai Wonderswan and Bandai Wonderswan Color are exceptions for one
# main reason: their consoles can often be played in a rotated style. To avoid trying to figure
# out which way we need to scale, we lie and say that the screen size for these consoles is square.
# This allows us to scale the image without needing to care about what orientation they are
# currently running in, which would have been impossible to deal with without special casing
# each and every little game... too much work for me. This only works because the resolution these
# handhelds run at aren't that far off from being square anyways.

if [ "$1" = "lynx" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="160"
  GAME_HEIGHT="160"
fi

if [ "$1" = "ws" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="224"
  GAME_HEIGHT="224"
fi

if [ "$1" = "wscolor" ]; then
  EMU_BIN="mednafen"
  GAME_WIDTH="224"
  GAME_HEIGHT="224"
fi

# Output the screen's width and height for debugging purposes.

echo "Game's screen width: $GAME_WIDTH"
echo "Game's screen height: $GAME_HEIGHT"

# Calculate the best scaling factor for the game in question by comparing the game's width
# and height to the size of the screen.

# Because of the way the scaler program worked up until now, we weren't previously
# able to scale to anything other than full integer factors (2x, 3x, etc...). This meant
# that we often were stuck with smaller scaling sizes than we'd like. The latest version
# of fscaler does allow you to use two passes on the image, with the second pass allowing
# downscaling by half. So to get some scaling factors such as 1.5x and 2.5x, you must be able
# to scale up to 3x and 5x first and then use the "half" scaling on the second pass.

# The SCALING_FACTOR variables here are calculated best scaling factors to the nearest
# full integer number. The SUBSCALING variables are used to figure out if we can go an
# extra 0.5x or not.

#By dividing the display size by the game size, we can find out the best scaling size to use.
SCALING_FACTOR_X=$(($DISPLAY_WIDTH/$GAME_WIDTH))

# However, when we're dealing with 0.5x magnifications, it doesn't really work. What we do
# instead is divide the same two numbers and see if there is a remainder. If the remainder is
# larger than half the game size, we know we can go with a non-integer scaling mode instead.

SUBSCALING_X=$(($DISPLAY_WIDTH%$GAME_WIDTH*2))
echo "Subscaling Factor X: $SUBSCALING_X"

# We could have set SUBSCALING to any number other than zero, really... but picking 5
# means it's easier to output the actual scaling factor for debugging.
if [ $SUBSCALING_X -gt $GAME_WIDTH ]; then
   SUBSCALING_X=5
else
   SUBSCALING_X=0
fi

# We've figured out the best scaling factor horizontally, now let's do the same calculations
# on the Y axis.

SCALING_FACTOR_Y=$(($DISPLAY_HEIGHT/$GAME_HEIGHT))
SUBSCALING_Y=$(($DISPLAY_HEIGHT%GAME_HEIGHT*2))
echo "Subscaling Factor Y: $SUBSCALING_Y"
if [ $SUBSCALING_Y -gt $GAME_HEIGHT ]; then
   SUBSCALING_Y=5
else
   SUBSCALING_Y=0
fi

echo "Optimal integer X scaling factor: $SCALING_FACTOR_X.$SUBSCALING_X"
echo "Optimal integer Y scaling factor: $SCALING_FACTOR_Y.$SUBSCALING_Y"

# Now that we've calculated the best scaling factors horizontally and vertically, we
# compare the two scaling factors and choose the smaller of the two. This is done
# because if we used the larger one instead, at least some part of the screen would
# be cut off. Chosing the smaller of the two ensures that we should always be able
# to see everything going on.

if [ "$SCALING_FACTOR_X" -le "$SCALING_FACTOR_Y" ]; then
   SCALING_FACTOR_T=`echo $SCALING_FACTOR_X`
   SUBSCALING_T=`echo $SUBSCALING_X`
else
   SCALING_FACTOR_T=`echo $SCALING_FACTOR_Y`
   SUBSCALING_T=`echo $SUBSCALING_Y`
fi
echo "Scaling game to $SCALING_FACTOR_T.$SUBSCALING_T"

# Now that we've decided the best scaling factor to use, let's run the emulator using
# the scaling type and scaling factor selected by the user. In some cases, certain
# scaling types and factors are not available due to how the scaler program works.

# As usual, MAME is somewhat of an exception because of required command line
# parameters. (Maybe these two are actually configurable in mame.ini, which
# would greatly simplify this part of the script... but I'm lazy. :)

# We run the scaler program and try to match up with what the user requested as
# best we can. The emulator being run is decided by the console being run, dealt
# with earlier in the script. Perhaps I'll add some more to handle more rarely used
# scaling factors, depending on what I see when I use 1080p.


if [ ! $GAMEEXCEPTION ]; then 
if [ "$2" = "fscaler" ]; then
  if [ "$SCALING_FACTOR_T" = 1 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f scale3x -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
      $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 2 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f near5 -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
      fscaler -f scale2x -t 999999 & $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 3 ]; then
    fscaler -f scale3x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 4 ]; then
    fscaler -f scale4x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 5 ]; then
    fscaler -f near5 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 6 ]; then
    fscaler -f scale3x -s scale2x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 7 ]; then
    fscaler -f scale3x -s scale2x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 8 ]; then
   fscaler -f scale4x -s scale2x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  killall fscaler
  exit
fi
if [ "$2" = "sscaler" ]; then
  if [ "$SCALING_FACTOR_T" = 1 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f scale3x -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
      $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 2 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f near5 -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
      fscaler -f s2xsai -t 999999 & $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 3 ]; then
    fscaler -f scale3x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 4 ]; then
    fscaler -f s2xsai -s scale2x -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 5 ]; then
    fscaler -f near5 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 6 ]; then
    fscaler -f scale3x -s s2xsai -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 7 ]; then
    fscaler -f scale3x -s s2xsai -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 8 ]; then
    fscaler -f scale4x -s s2xsai -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  killall fscaler
  exit
else
  if [ "$SCALING_FACTOR_T" = 1 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f near3 -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
    $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 2 ]; then
    if [ "$SUBSCALING_T" = 5 ]; then
      fscaler -f near5 -s half -t 999999 & $EMU_BIN "$GAMENAME"
    else
      fscaler -f near2 -t 999999 & $EMU_BIN "$GAMENAME"
    fi
  fi
  if [ "$SCALING_FACTOR_T" = 3 ]; then
    fscaler -f near3 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 4 ]; then
    fscaler -f near4 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 5 ]; then
    fscaler -f near5 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 6 ]; then
    fscaler -f near3 -s near2 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 7 ]; then
    fscaler -f near3 -s near2 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  if [ "$SCALING_FACTOR_T" = 8 ]; then
    fscaler -f near4 -s near2 -t 999999 & $EMU_BIN "$GAMENAME"
  fi
  # Once the emulator is done running, we kill off the scaler program.
  killall fscaler
fi
else
  if [ $MANAHACK ]; then
   cp ~/.snes9x/snes9x-mana.xml ~/.snes9x/snes9x.xml
   $EMU_BIN "$GAMENAME"
   cp ~/.snes9x/snes9x-orig.xml ~/.snes9x/snes9x.xml
  fi
fi

# Thank you, and good night everybody!
# *crowd cheers*
